<?php
/* Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Controller\TAXII;

use Ramsey\Uuid\Uuid;
use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Providers\QueryProvider;
use App\Security\Voter\FsrVoter;
use App\Constants\CacheConstants;
use App\Constants\ErrorConstants;
use App\Entity\Query\SystemQuery;
use App\Entity\Custom\ThreatIntel;
use App\Entity\Settings\SystemSettings;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use App\Providers\PermissionsDictionaryBuilder;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\Serializer\SerializerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use ApiPlatform\Core\DataProvider\CollectionDataProviderInterface;
use ApiPlatform\Core\Serializer\SerializerContextBuilderInterface;
use App\Service\UtilityService;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;

class TaxiiApiController extends AbstractController
{

    public const THREAT_FEED_MODULE_NAME = 'threat_intel_feeds';
    private $csvDir;

    public function __construct(
        protected EntityManagerInterface $entityManager,
        protected SerializerInterface $serializer,
        protected SerializerContextBuilderInterface $serializerContextBuilder,
        protected CollectionDataProviderInterface $collectionDataProvider,
        protected FsrVoter $fsrVoter,
        protected QueryProvider $queryProvider,
        protected ContainerInterface $containers,
        protected LoggerInterface $logger,
        protected UtilityService $utilityService
    ) {
        $this->csvDir = $this->containers->getParameter('kernel.project_dir') . '/workspace/';
    }

    /**
     * @Route("api/taxii/1/taxii", name="app_api_root", methods={"GET"})
     */
    public function getTaxiiRootApi()
    {
        $module = self::THREAT_FEED_MODULE_NAME;
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module)) {
            throw new AccessDeniedException();
        }

        $wfClient = $this->containers->get("app.proxy.client.workflow");
        $headers = array(
            "Content-type" => "application/json",
        );
        $workflowUri = $this->containers->getParameter('workflow.uri');
        $workflowRequests = new GuzzleRequest('get', $workflowUri . '/api/dynamic-variable/?format=json&name=Server_fqhn', $headers);
        $system_host = 'localhost';
        try {
            $workflowResponse = $wfClient->send($workflowRequests);
            $workflowResponseData = json_decode($workflowResponse->getBody()->getContents(), true);
            if ($workflowResponseData["hydra:totalItems"] > 0) {
                $system_host = $workflowResponseData["hydra:member"][0]["value"];
            }
        } catch (BadRequestHttpException | \Exception $error) {
            $this->logger->error($error->getMessage());
        }

        $responseObject = [
            "api_roots" => [
                "https://$system_host/api/taxii/1/"
            ],
            "contact" => "fsr-integrations@fortinet.com",
            "default" => "https://$system_host/api/taxii/1/",
            "description" => "TAXII 2.1 Server",
            "title" => "FortiSOAR Taxii 2.1 Server"
        ];
        $response = new Response();
        $response->headers->set('Content-Type', 'application/taxii+json;version=2.1');
        $response->setContent(json_encode($responseObject));
        return $response;
    }

    /**
     * @Route("api/taxii/1/", name="app_api_root_info", methods={"GET"})
     */
    public function getTaxiiRootApiInfo()
    {
        $module = self::THREAT_FEED_MODULE_NAME;
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module)) {
            throw new AccessDeniedException();
        }
        $responseObject = [
            "max_content_length" => 1000000,
            "title" => "TAXII feeds",
            "versions" => [
                "taxii-2.1"
            ]
        ];
        $response = new Response();
        $response->headers->set('Content-Type', 'application/taxii+json;version=2.1');
        $response->setContent(json_encode($responseObject));
        return $response;
    }

    /**
     * @Route("api/taxii/1/collections", name="app_api_feeds_collections", methods={"GET"})
     */
    public function getFeedsCollection(Request $request)
    {
        $isTimSkuEnabled = $this->utilityService->isTimSkuEnabled();
        $module = self::THREAT_FEED_MODULE_NAME;
        $systemSettingRepo = $this->entityManager->getRepository(SystemSettings::class);
        $taxiiSettingObj = $systemSettingRepo->findOneBy(['name' => "TAXII Server Details"]);
        if (!is_null($taxiiSettingObj) && !$taxiiSettingObj->getPublicValues()['status']['enabled']) {
            throw new BadRequestHttpException(ErrorConstants::TAXII_SERVER_DISABLED);
        }
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module)) {
            throw new AccessDeniedException("Access Denied");
        }

        // Keeping default value as 100 for query result, for backward compatability.
        // Limiting the count of collection being return for non SKU
        // This needs to be confirmed from PM too

        $queryLimit = $isTimSkuEnabled ? $request->get('$limit', 100) : AppConstants::TOTAL_RESTRICTED_COLLECTIONS_COUNT;
        $queryPage = $isTimSkuEnabled ? $request->get('$page', 1) : AppConstants::TOTAL_RESTRICTED_FEEDS_PAGE;

        $systemQueryRepo = $this->entityManager->getRepository(SystemQuery::class);
        $feedCollectionsQueryBuilder = $systemQueryRepo->createQueryBuilder('o')
            ->join('o.models', 'm')
            ->andWhere('m.type = :types')
            ->setParameter('types', $module);

        if (!$isTimSkuEnabled) {
            $feedCollectionsQueryBuilder->setFirstResult($queryPage);
            $feedCollectionsQueryBuilder->setMaxResults($queryLimit);
        }

        $feedCollections = $feedCollectionsQueryBuilder->getQuery()->getResult();

        $collectionList = [];
        $collectionList['collections'] = [];
        if (!empty($feedCollections)) {
            foreach ($feedCollections as $filter) {
                $collectionList['collections'][] = [
                    "name" => $filter->getName(),
                    "id" => $filter->getUuid(),
                    "uuid" => $filter->getUuid()
                ];
            }
        }
        $response = new Response();
        $response->headers->set('Content-Type', 'application/taxii+json;version=2.1');
        $response->setContent(json_encode($collectionList));
        return $response;
    }


    private function getRequestContext($request, $attributes, $selectFields = null)
    {
        $request->overrideGlobals();
        $context = $this->serializerContextBuilder->createFromRequest($request, true, $attributes);
        $context["relationships"] =  false;
        $context["ignore_authorization"] =  true;
        $context["ignore_field_authorization"] =  true;
        $context["ignore_relationship_authorization"] =  true;
        $context["displayNameOnly"] =  true;
        if ($selectFields) {
            $context['groups'] = explode(',', $selectFields);
        }
        $context['ignored_attributes'] = AppConstants::TRAITS_ATTRIBUTES;
        return $context;
    }

    private function identifyQueryLimitAndPage($request)
    {
        $isTimSkuEnabled = $this->utilityService->isTimSkuEnabled();
        $queryLimit = $request->get('$limit', 30);
        $queryPage = $request->get('$page', 1);
        if (!$isTimSkuEnabled) {
            $maxQueryLimit = $this->utilityService->getRequiredLicenseDetails(CacheConstants::TIM_SUPPORTED_QUERY_LIMIT);
            // For non SKU User, we will only be seding max batch of 100 records.
            $maxQueryPage = ceil($maxQueryLimit / $queryPage);
            $queryLimit = $queryLimit > $maxQueryLimit ? $maxQueryLimit : $queryLimit;
            $queryPage = $queryPage > $maxQueryPage ? $maxQueryPage : $queryPage;
        }
        return [$queryLimit, $queryPage];
    }

    // Support Feed Listing
    // Support Filter Feed by Feed ID, Collection Name
    // Support Filter Feed by combination of Feed ID and Collection Name
    // All Response return are in Taxii 2.1 format

    /**
     * @Route("api/taxii/1/collections/{collectionId}/{objects?}/{objectId?}", name="app_api_feeds_export", requirements={"objectId"=".+"}, methods={"GET"})
     */
    public function getFeeds(Request $request, $collectionId, $objects = 'objects', $objectId = null)
    {
        $isTimSkuEnabled = $this->utilityService->isTimSkuEnabled();
        $accept = $request->headers->get('Accept', 'application/json');
        $format = $_GET['$format'];
        $selectFields = $_GET['$__selectFields'];
        $systemSettingRepo = $this->entityManager->getRepository(SystemSettings::class);
        $taxiiSettingObj = $systemSettingRepo->findOneBy(['name' => "TAXII Server Details"]);
        if (!is_null($taxiiSettingObj) && !$taxiiSettingObj->getPublicValues()['status']['enabled']) {
            throw new BadRequestHttpException(ErrorConstants::TAXII_SERVER_DISABLED);
        }
        $module = self::THREAT_FEED_MODULE_NAME;
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module)) {
            throw new AccessDeniedException("Access Denied");
        }
        $resourceClass = ThreatIntel::class;

        // This API is consumed by third party products so rather than pagination support for non SKU
        // It will be easy to give total of 100 records in single page
        // Customer still can create multiple collections and fetch data using different set of API
        // Need to validate if we should be restricting making limited filters for non SKU customers
        [$queryLimit, $queryPage] = $this->identifyQueryLimitAndPage($request);
        $partial = $request->get('$partial', false);
        $attributes = [];
        $attributes['filters'] = ['$limit' => $queryLimit, '$page' => $queryPage, '$partial' => $partial];
        $attributes['resource_class'] = $resourceClass;
        $attributes['collection_operation_name'] = "get";
        $systemQueryRepo = $this->entityManager->getRepository(SystemQuery::class);
        $savedFilter = $systemQueryRepo->findOneBy(['uuid' => $collectionId]);
        if (is_null($savedFilter)) {
            throw new BadRequestHttpException(sprintf('Collection ID %s not found', $collectionId));
        }

        if (is_null($objects)) {
            $response = new Response();
            $response->headers->set('Content-Type', 'application/taxii+json;version=2.1');
            $collectionDetails = [
                "name" => $savedFilter->getName(),
                "id" => $savedFilter->getUuid(),
                "uuid" => $savedFilter->getUuid()
            ];
            return $response->setContent(json_encode($collectionDetails));
        }

        $filterObj = $savedFilter->getQuery();
        foreach ($filterObj['filters'] as $index => $filterValue) {
            if (is_array($filterValue['value'])) {
                $filterObj['filters'][$index]['value'] = [$filterValue['value'][0]['@id']];
            }
        }
        if (!is_null($objectId)) {
            $filterObj['filters'][$index + 1] = [
                "field" => "uuid",
                "operator" => "like",
                "type" => "primitive",
                "value" => '%' . $objectId . '%'
            ];
        }
        $this->queryProvider->setInitialQuery($filterObj);
        $context = $this->getRequestContext($request, $attributes, $selectFields);
        $feedByCollection = $this->collectionDataProvider->getCollection($resourceClass, $attributes['collection_operation_name'], $attributes);
        if (empty($feedByCollection)) {
            throw new NotFoundHttpException();
        }
        $normalized_data = $this->serializer->normalize($feedByCollection, 'jsonld', $context);
        unset($normalized_data['@context']);
        unset($normalized_data['@id']);
        unset($normalized_data['@type']);
        unset($normalized_data['hydra:totalItems']);
        if (is_array($normalized_data["hydra:view"]) && isset($normalized_data["hydra:view"]["hydra:next"])) {
            $normalized_data['more'] = true;
            $normalized_data['next'] = $queryPage + 1;
        }
        $normalized_data['objects'] = $normalized_data['hydra:member'];
        unset($normalized_data['hydra:member']);
        unset($normalized_data["hydra:view"]);
        $response = new Response();
        if ($accept == 'text/csv' || $format == 'csv') {
            if (!empty($normalized_data['objects'])) {
                $filePath = $this->csvDir . $savedFilter->getName() . '-' . Uuid::uuid4()->toString() . '.csv';
                $this->createCSV($normalized_data['objects'], $filePath);
                $fileContent = file_get_contents($filePath);
                $response->headers->set('Content-Type', 'text/csv');
                $response->headers->set('Content-Disposition', 'attachment; filename="' . basename($filePath) . '"');
                $response->headers->set('filename', basename($filePath));
                $response->headers->set('Content-Length', filesize($filePath));
                $response->setContent($fileContent);
                unlink($filePath);
                return $response;
            }
            return new NotFoundHttpException('No record found matching the given objects ID');
        }
        $response->headers->set('Content-Type', 'application/taxii+json;version=2.1');
        $response->setContent(json_encode($normalized_data));

        return $response;
    }

    protected function createCSV($normalized_object, $filePath)
    {
        $headers = [];
        foreach ($normalized_object[0] as $key => $value) {
            if ($key == "@id" || $key == "@type") {
                continue;
            }
            $headers[] = $key;
        }
        $csvData[0] = $headers;
        foreach ($normalized_object as $index => $objects) {
            $data = [];
            foreach ($objects as $key => $value) {
                if (!is_array($value)) {
                    if ($key == "@id" || $key == "@type") {
                        continue;
                    }
                    $data[] = $value;
                } else {
                    $data[] = implode(',', $value);
                }
            }
            $csvData[(int)$index + 1] = $data;
        }

        $fp = fopen($filePath, 'wb');
        foreach ($csvData as $line) {
            fputcsv($fp, $line, ',');
        }
        fclose($fp);
    }
}
