<?php
/* Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Constants\CacheConstants;
use App\Constants\ErrorConstants;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class SKUService
{
    protected $dasClient;
    protected $dasUri;
    private $headers = array("Content-type" => "application/json");

    public function __construct(
        private LoggerInterface $logger,
        protected ContainerInterface $containers,
        protected TagAwareCacheInterface $cache
    ) {
        $this->dasClient = $this->containers->get("app.proxy.client.das");
        $this->dasUri = $this->containers->getParameter('das.uri');
    }

    public function getDetailsFromDas()
    {
        $timEnabled = false;
        $isDistributedInstance = false;

        $instanceRole = AppConstants::MASTER_ROLE_NAME;
        $supportedIngestionLimit = AppConstants::TOTAL_INGESTION_FEEDS_LIMIT;
        $supportedQueryLimit = AppConstants::TOTAL_RESTRICTED_FEEDS_COUNT;

        $dasRequests = new GuzzleRequest('get', $this->dasUri . '/license/?param=license_details', $this->headers);
        try {
            $dasResponse = $this->dasClient->send($dasRequests);
            $dasResponseData = $dasResponse->getBody()->getContents();
            $jsonDecodeResponse = json_decode($dasResponseData, true);
            $resCode = $dasResponse->getStatusCode();
            if ($resCode < 300) {
                $isDistributedInstance = $jsonDecodeResponse['details']['is_distributed'] ?? false;
                $instanceRole = $isDistributedInstance ? $jsonDecodeResponse['details']['role'][0] : null;

                // Contracts key is always expected
                $timEnabled = $jsonDecodeResponse['details']['contracts']['has_tim_subscription'] ?? false;
                $supportedIngestionLimit = $jsonDecodeResponse['details']['contracts']['tim_supported_ingestion_limit'] ?? AppConstants::TOTAL_INGESTION_FEEDS_LIMIT;
                $supportedQueryLimit = $jsonDecodeResponse['details']['contracts']['tim_supported_query_limit'] ?? AppConstants::TOTAL_RESTRICTED_FEEDS_COUNT;

                $this->setCache($isDistributedInstance, $instanceRole, $timEnabled, $supportedIngestionLimit, $supportedQueryLimit);
                return [$isDistributedInstance, $instanceRole, $timEnabled, $supportedIngestionLimit, $supportedQueryLimit];
            }
        } catch (BadRequestHttpException | \Exception $error) {
            $this->logger->info($error->getMessage());
        }
        return [$isDistributedInstance, $instanceRole, $timEnabled, $supportedIngestionLimit, $supportedQueryLimit];
    }

    private function setCache($isDistributedInstance, $instanceRole, $timEnabled, $supportedIngestionLimit, $supportedQueryLimit)
    {
        $item = $this->cache->getItem(CacheConstants::IS_DISTRIBUTED_INSTANCE);
        $item->set($isDistributedInstance);
        $item->tag(CacheConstants::MSSP_TYPE_ROLE);
        $this->cache->save($item);

        $item = $this->cache->getItem(CacheConstants::INSTANCE_ROLE);
        $item->set($instanceRole);
        $item->tag(CacheConstants::MSSP_TYPE_ROLE);
        $this->cache->save($item);

        $item = $this->cache->getItem(CacheConstants::IS_TIM_ENABLED);
        $item->set($timEnabled);
        $item->tag(CacheConstants::IS_TIM_ENABLED);
        $this->cache->save($item);

        $item = $this->cache->getItem(CacheConstants::TIM_SUPPORTED_INGESTION_LIMIT);
        $item->set($supportedIngestionLimit);
        $item->tag(CacheConstants::TIM_SUPPORTED_INGESTION_LIMIT);
        $this->cache->save($item);

        $item = $this->cache->getItem(CacheConstants::TIM_SUPPORTED_QUERY_LIMIT);
        $item->set($supportedQueryLimit);
        $item->tag(CacheConstants::TIM_SUPPORTED_QUERY_LIMIT);
        $this->cache->save($item);
    }

    public static function computeHash(string $secret_key, int $usedLimit, \Datetime $now)
    {
        $jsonData = [
            'date' => $now->format('U'),
            'feedUsedLimit' => AppConstants::STATIC_STRING . strval($usedLimit)
        ];
        $data = json_encode($jsonData);
        return $secret_key . "#" . hash("sha512", $secret_key . $data);
    }

    public static function validateHash(string $checksum, int $usedLimit, $modifiedAt)
    {
        $jsonData = [
            'date' => $modifiedAt->format('U'),
            'feedUsedLimit' => AppConstants::STATIC_STRING . strval($usedLimit)
        ];
        $data = json_encode($jsonData);
        if (hash("sha512", substr($checksum, 0, strpos($checksum, "#")) . $data) == substr($checksum, strpos($checksum, "#") + 1)) {
            return true;
        }
        return false;
    }

    public static function identifyCurrentDayLimit(\DateTime $modifiedAt, int $usedLimit)
    {
        $today = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $diff = ($today->setTime(00, 00))->diff($modifiedAt->setTime(00, 00));
        $diffDays = (int)$diff->format("%R%a");
        if ($diffDays != 0) {
            return 0;
        }
        return $usedLimit;
    }


    public static function validateModifyDate(\DateTime $lastCreatedFeed, \DateTime $modifiedAt)
    {
        $dateDiff = $lastCreatedFeed->diff($modifiedAt);
        if ($dateDiff->i > AppConstants::MAX_TIME_GAP_ALLOWED) {
            throw new BadRequestHttpException(ErrorConstants::LICENSE_TAMPERING_DETECTED);
        }
    }
}
